/****************************************************************************
 *  Copyright (C) 2017-2018 Kron Technologies Inc <http://www.krontech.ca>. *
 *                                                                          *
 *  This program is free software: you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation, either version 3 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
 ****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <glib.h>
#include <dbus/dbus-glib.h>

#include "jsmn.h"
#include "dbus-json.h"
#include "api/cam-rpc.h"

#define OPT_FLAG_RPC    (1<<0)
#define OPT_FLAG_CGI    (1<<1)
#define OPT_FLAG_GET    (1<<2)

/* Standard JSON-RPC Error Codes. */
#define JSONRPC_ERR_PARSE_ERROR         (-32700)
#define JSONRPC_ERR_INVALID_REQUEST     (-32600)
#define JSONRPC_ERR_METHOD_NOT_FOUND    (-32601)
#define JSONRPC_ERR_INVALID_PARAMETERS  (-32602)
#define JSONRPC_ERR_INTERNAL_ERROR      (-32603)
#define JSONRPC_ERR_SERVER_ERROR        (-32604)

static void
handle_error(int code, const char *message, unsigned long flags)
{
    if (flags & OPT_FLAG_RPC) {
        fputs("{\n", stdout);
        fprintf(stdout, "%*.s\"jsonrpc\": \"2.0\",\n", JSON_TAB_SIZE, "");
        fprintf(stdout, "%*.s\"id\": null,\n", JSON_TAB_SIZE, "");
        fprintf(stdout, "%*.s\"error\": {\n", JSON_TAB_SIZE, "");
        fprintf(stdout, "%*.s\"code\": %d,\n", 2 * JSON_TAB_SIZE, "", code);
        fprintf(stdout, "%*.s\"message\": ", 2 * JSON_TAB_SIZE, "");
        json_printf_utf8(stdout, message);
        fprintf(stdout, "\n%*.s}\n", JSON_TAB_SIZE, "");
        fputs("}\n", stdout);
    }
    else if (flags & OPT_FLAG_CGI) {
        fprintf(stdout, "Content-type: application/json\n");
        switch (code) {
            case JSONRPC_ERR_PARSE_ERROR:
            case JSONRPC_ERR_INVALID_REQUEST:
            case JSONRPC_ERR_INVALID_PARAMETERS:
                fprintf(stdout, "Status: 400 Bad Request\n");
                break;

            case JSONRPC_ERR_METHOD_NOT_FOUND:
                fprintf(stdout, "Status: 404 Not Found\n");
                break;
            
            case JSONRPC_ERR_INTERNAL_ERROR:
            case JSONRPC_ERR_SERVER_ERROR:
                fprintf(stdout, "Status: 500 Internal Server Error\n");
        }
        fputs("\n{\n", stdout);
        fprintf(stdout, "%*.s\"error\": ", JSON_TAB_SIZE, "");
        json_printf_utf8(stdout, message);
        fprintf(stdout, "\n}\n");
    }
    else {
        fprintf(stderr, "RPC Call Failed: %s\n", message);
    }
    exit(EXIT_FAILURE);
}

/* Put some upper bounds on the amount of JSON we're willing to parse. */
#define JSON_MAX_BUF    4096
#define JSON_MAX_TOK    128

static char js[JSON_MAX_BUF];

/* Recursive helper to get the real size, in tokens, of a non-trivial token. */
static inline int
json_token_size(jsmntok_t *start)
{
    /* Recursively count objects. */
    if (start->type == JSMN_OBJECT) {
        int i;
        jsmntok_t *tok = start+1;
        for (i = 0; i < start->size; i++) {
            tok++;
            tok += json_token_size(tok);
        }
        return tok - start;
    }
    /* Recursively count arrays. */
    if (start->type == JSMN_ARRAY) {
        int i;
        jsmntok_t *tok = start+1;
        for (i = 0; i < start->size; i++) {
            tok += json_token_size(tok);
        }
        return tok - start;
    }
    /* Everything else has size 1. */
    return 1;
}

/* Parse a simple JSON array where all members are of the same type. */
static gpointer
json_parse_array(jsmntok_t *tokens, GType *ptype, unsigned long flags)
{
    int children, i = 1;

    /* String cases - use a GPtrArray */
    if (tokens[1].type == JSMN_STRING) {
        GPtrArray *arr = g_ptr_array_sized_new(tokens->size);
        if (!arr) {
            handle_error(JSONRPC_ERR_INTERNAL_ERROR, "Internal error", flags);
        }
        g_ptr_array_set_free_func(arr, g_free);
        *ptype = dbus_g_type_get_collection("GPtrArray", G_TYPE_STRING);

        /* Add all strings to the pointer array.  */
        for (children = 0; children < tokens->size; children++) {
            jsmntok_t *tok = &tokens[i];
            if (tok->type == JSMN_STRING) {
                char *value = &js[tok->start];
                value[tok->end - tok->start] = '\0';
                g_ptr_array_add(arr, g_strdup(value));
            }
            i += json_token_size(tok);
        }
        return arr;
    }
    else if (tokens[1].type != JSMN_PRIMITIVE) {
        /* Arrays of complex types are not supported. */
        return NULL;
    }
    /* If the first token is 'true', 'false', or 'null' then build an array of booleans. */
    else if (memchr("tfn", js[tokens[1].start], 3) != NULL) {
        GArray *arr = g_array_sized_new(FALSE, TRUE, sizeof(gboolean), tokens->size);
        if (!arr) {
            handle_error(JSONRPC_ERR_INTERNAL_ERROR, "Internal error", flags);
        }
        *ptype = dbus_g_type_get_collection("GArray", G_TYPE_BOOLEAN);
        g_array_set_size(arr, tokens->size);
        
        /* Add all doubles to the pointer array. */
        for (children = 0; children < tokens->size; children++) {
            jsmntok_t *tok = &tokens[i];
            if (tok->type == JSMN_PRIMITIVE) {
                g_array_index(arr, gboolean, children) = (js[tok->start] == 't');
            }
            i += json_token_size(tok);
        }
    }
    /*
     * Otherwise, the array could contain integers or floats, choose
     * the lowest common denominator and build an array of doubles.
     */
    else {
        GArray *arr = g_array_sized_new(FALSE, TRUE, sizeof(gdouble), tokens->size);
        if (!arr) {
            handle_error(JSONRPC_ERR_INTERNAL_ERROR, "Internal error", flags);
        }
        *ptype = dbus_g_type_get_collection("GArray", G_TYPE_DOUBLE);
        g_array_set_size(arr, tokens->size);

        /* Add all doubles to the pointer array. */
        for (children = 0; children < tokens->size; children++) {
            jsmntok_t *tok = &tokens[i];
            if (tok->type == JSMN_PRIMITIVE) {
                char *value = &js[tok->start];
                value[tok->end - tok->start] = '\0';
                g_array_index(arr, gdouble, children) = strtod(value, NULL);
            }
            i += json_token_size(tok);
        }
        return arr;
    }
}

static GHashTable *
json_parse_object(jsmntok_t *tokens, unsigned long flags)
{
    int children, i = 1;
    GHashTable *h = cam_dbus_dict_new();
    if (!h) {
        handle_error(JSONRPC_ERR_INTERNAL_ERROR, "Internal error", flags);
    }

    /* Parse the tokens making up this JSON object. */
    for (children = 0; children < tokens[0].size; children++) {
        jsmntok_t   *tok;
        char        *name;
        char        *value;

        /* To be a valid JSON object, the first token must be a string. */
        if (tokens[i].type != JSMN_STRING) {
            handle_error(JSONRPC_ERR_PARSE_ERROR, "Invalid JSON", flags);
        }
        name = &js[tokens[i].start];
        name[tokens[i].end - tokens[i].start] = '\0';
        i++;

        /* The following token can be any simple type. */
        tok = &tokens[i];
        value = &js[tok->start];
        value[tok->end - tok->start] = '\0';
        i += json_token_size(tok);

        if (tok->type == JSMN_STRING) {
            /* TODO: Deal with escaped UTF-8. This is probably a security hole. */
            cam_dbus_dict_add_printf(h, name, "%s", value);
        }
        else if (tok->type == JSMN_OBJECT) {
            GHashTable *obj = json_parse_object(tok, flags);
            if (!obj) {
                handle_error(JSONRPC_ERR_INTERNAL_ERROR, "Internal error", flags);
            }
            cam_dbus_dict_take_boxed(h, name, CAM_DBUS_HASH_MAP, obj);
        }
        else if (tok->type == JSMN_ARRAY) {
            GType arraytype;
            gpointer *array = json_parse_array(tok, &arraytype, flags);
            if (!array) {
                handle_error(JSONRPC_ERR_INTERNAL_ERROR, "Internal error", flags);
            }
            cam_dbus_dict_take_boxed(h, name, arraytype, array);
        }
        else if (tok->type != JSMN_PRIMITIVE) {
            /* Ignore other nested types. */
            /* I think the only possible types left are arrays. */
            continue;
        }
        /* Break it down by primitive types. */
        else if (*value == 't') {
            cam_dbus_dict_add_boolean(h, name, 1);
        } else if (*value == 'f') {
            cam_dbus_dict_add_boolean(h, name, 0);
        } else if (*value == 'n') {
            /* TODO: Explicit NULL types and other magic voodoo? */
        }
        /* Below here, we have some kind of numeric type. */
        else if (strcspn(value, ".eE") != (tok->end - tok->start)) {
            char *e;
            double x = strtod(value, &e);
            if (*e != 0) continue; /* TODO: throw an "invalid JSON"? */
            cam_dbus_dict_add_float(h, name, x);
        } else if (*value == '-') {
            /* Use the type 'long long' for negative values. */
            char *e;
            long long x = strtoll(value, &e, 0);
            if (*e != 0) continue; /* TODO: throw an "invalid JSON"? */
            cam_dbus_dict_add_int(h, name, x);
        } else {
            char *e;
            unsigned long long x = strtoull(value, &e, 0);
            if (*e != 0) continue; /* TODO: throw an "invalid JSON"? */
            cam_dbus_dict_add_uint(h, name, x);
        }
    }

    /* Success */
    return h;
}

/* Parse JSON args, or return NULL if no args were provided. */
static void *
json_parse(FILE *fp, GType *ptype, unsigned long flags)
{
    jsmn_parser parser;
    jsmntok_t tokens[JSON_MAX_TOK]; /* Should be enough for any crazy API thing. */
    size_t  jslen;
    int     num;

    /* Read the file for JSON data. */
    jslen = fread(js, 1, sizeof(js), fp);
    if (!jslen) {
        /* If we just get an EOF then assume no args. */
        if (feof(fp)) return NULL;
        /* Otherwise, report a parsing error. */
        handle_error(JSONRPC_ERR_PARSE_ERROR, "Invalid JSON", flags);
    }

    /* If there is more input, then our buffer is too small. */
    if (!feof(fp)) {
        /* We really want a 'request too large' error for CGI mode. */
        handle_error(JSONRPC_ERR_PARSE_ERROR, "Invalid JSON", flags);
    }

    jsmn_init(&parser);
    num = jsmn_parse(&parser, js, jslen, tokens, JSON_MAX_TOK);
    if (num < 0) {
        /* We really want a 'request too large' error for CGI mode. */
        handle_error(JSONRPC_ERR_PARSE_ERROR, "Invalid JSON", flags);
    }

    /* The only encoding we support for now is the JSON object */
    if (tokens[0].type == JSMN_OBJECT) {
        *ptype = CAM_DBUS_HASH_MAP;
        return json_parse_object(tokens, flags);
    }
    if (tokens[0].type == JSMN_ARRAY) {
        return json_parse_array(tokens, ptype, flags);
    }

    /* Not any kind of input that we can make sense of. */
    return NULL;
}

static void
usage(FILE *fp, int argc, char * const argv[])
{
    fprintf(fp, "Usage: %s [options] [METHOD [PARAMS]]\n\n", argv[0]);

    fprintf(fp, "Make a DBus call to the Chronos camera daemon, and translate\n");
    fprintf(fp, "the result into JSON.\n\n");

    fprintf(fp, "In normal and JSON-RPC mode, the RPC method is provided as the\n");
    fprintf(fp, "first positional argument. If the method takes arguments, they\n");
    fprintf(fp, "will be parsed from the optional PARAMS file (or \'-\' to read\n");
    fprintf(fp, "from stdin).\n\n");

    fprintf(fp, "In CGI mode, the PATH_INFO environment variable is parsed for\n");
    fprintf(fp, "the RPC method to call and arguments will be parsed from stdin.\n\n");
    
    fprintf(fp, "In get mode, the \'get\' method will be called to retrieive\n");
    fprintf(fp, "parameters from the DBus interface. The names of the parameters\n");
    fprintf(fp, "to retrieve are provided as the positional arguments to %s.\n\n", argv[0]);

    fprintf(fp, "options:\n");
    fprintf(fp, "\t-r, --rpc     encode the results in JSON-RPC format\n");
    fprintf(fp, "\t-c, --cgi     encode the results in CGI/1.0 format\n");
    fprintf(fp, "\t-g, --get     get paramereters from the DBus interface\n");
    fprintf(fp, "\t-n, --control connect to the control DBus interface\n");
    fprintf(fp, "\t-v, --video   connect to the video DBus interface\n");
    fprintf(fp, "\t-h, --help    display this help and exit\n");
}

int
main(int argc, char * const argv[])
{
    DBusGConnection* bus;
    DBusGProxy* proxy;
    GHashTable *h;
    void *params = NULL;
    GType paramtype = G_TYPE_INVALID;
    GError* error = NULL;
    gboolean okay;
    const char *service = CAM_DBUS_CONTROL_SERVICE;
    const char *path = CAM_DBUS_CONTROL_PATH;
    const char *iface = CAM_DBUS_CONTROL_INTERFACE;
    const char *method;
    unsigned long flags = 0;
    
    /* Option Parsing */
    const char *short_options = "rcgvnh";
    const struct option long_options[] = {
        {"rpc",     no_argument,    0, 'r'},
        {"cgi",     no_argument,    0, 'c'},
        {"get",     no_argument,    0, 'g'},
        {"control", no_argument,    0, 'n'},
        {"video",   no_argument,    0, 'v'},
        {"help",    no_argument,    0, 'h'},
        {0, 0, 0, 0}
    };
    int c;
    optind = 0;
    while ((c = getopt_long(argc, argv, short_options, long_options, NULL)) > 0) {
        switch (c) {
            case 'r':
                flags |= OPT_FLAG_RPC;
                break;
            
            case 'c':
                flags |= OPT_FLAG_CGI;
                break;
            
            case 'g':
                flags |= OPT_FLAG_GET;
                break;
            
            case 'v':
                service = CAM_DBUS_VIDEO_SERVICE;
                path = CAM_DBUS_VIDEO_PATH;
                iface = CAM_DBUS_VIDEO_INTERFACE;
                break;

            case 'n':
                service = CAM_DBUS_CONTROL_SERVICE;
                path = CAM_DBUS_CONTROL_PATH;
                iface = CAM_DBUS_CONTROL_INTERFACE;
                break;

            case 'h':
                usage(stdout, argc, argv);
                return EXIT_SUCCESS;
            case '?':
            default:
                return EXIT_FAILURE;
        }
    }
    g_type_init();

    /*
     * If getting parameters, the method is 'get' and the positional
     * arguments are the names of the parameters to get.
     */
    if (flags & OPT_FLAG_GET) {
        method = "get";
        params = g_ptr_array_sized_new(argc - optind);
        paramtype = dbus_g_type_get_collection("GPtrArray", G_TYPE_STRING);
        if (!params) {
            handle_error(JSONRPC_ERR_INTERNAL_ERROR, "Internal error", flags);
        }
        g_ptr_array_set_free_func(params, g_free);
        while (optind < argc) {
            g_ptr_array_add(params, g_strdup(argv[optind++]));
        }
    }
    /* If CGI, get the requested method from the PATH_INFO variable. */
    else if (flags & OPT_FLAG_CGI) {
        method = getenv("PATH_INFO");
        if (!method) {
            fprintf(stderr, "Missing variable: PATH_INFO\n");
            fprintf(stderr, "Try '%s --help' for more information.\n", argv[0]);
            return EXIT_FAILURE;
        }
        /* Stip leading slashes from PATH_INFO */
        while (*method == '/') method++;

        /* Attempt to parse the method arguments from stdin. */
        params = json_parse(stdin, &paramtype, flags);
    }
    /* Otherwise, the method name is passed in via the command line. */
    else if (optind >= argc) {
        fprintf(stderr, "Missing argument: METHOD\n");
        fprintf(stderr, "Try '%s --help' for more information.\n", argv[0]);
        return EXIT_FAILURE;
    }
    else {
        method = argv[optind++];

        /* If yet another parameter is present, it may provide a source file for
        * the RPC request parameters, or it may be '-' to read paramers from stdin.
        */
        if (optind < argc) {
            const char *filename = argv[optind++];
            FILE *fp = strcmp(filename, "-") ? fopen(filename, "r") : stdin;
            if (!fp) {
                fprintf(stderr, "Failed to open '%s' for reading: %s\n", filename, strerror(errno));
                return EXIT_FAILURE;
            }
            params = json_parse(fp, &paramtype, flags);
            fclose(fp);
        }
    }
    
    /* Initialize the DBus system. */
    bus = dbus_g_bus_get(DBUS_BUS_SYSTEM, &error);
    if (error != NULL) {
        handle_error(-32603, "Internal error", flags);
    }
    proxy = dbus_g_proxy_new_for_name(bus, service, path, iface);
    if (proxy == NULL) {
        handle_error(-32603, "Internal error", flags);
    }
    if (params) {
        okay = dbus_g_proxy_call(proxy, method, &error,
                paramtype, params, G_TYPE_INVALID,
                CAM_DBUS_HASH_MAP, &h, G_TYPE_INVALID);
    } else {
        okay = dbus_g_proxy_call(proxy, method, &error, G_TYPE_INVALID,
                CAM_DBUS_HASH_MAP, &h, G_TYPE_INVALID);
    }
    if (!okay) {
        handle_error(error->code, error->message, flags);
    }

    /* Output CGI header data */
    if (flags & OPT_FLAG_CGI) {
        fprintf(stdout, "Content-type: application/json\n");
        fprintf(stdout, "\n");
    }

    /* Output format: JSON-RPC */
    if (flags & OPT_FLAG_RPC) {
        fputs("{\n", stdout);
        fprintf(stdout, "%*.s\"jsonrpc\": \"2.0\",\n", JSON_TAB_SIZE, "");
        fprintf(stdout, "%*.s\"id\": null,\n", JSON_TAB_SIZE, "");
        fprintf(stdout, "%*.s\"result\": ", JSON_TAB_SIZE, "");
        json_printf_dict(stdout, h, 1);
        fputs("\n}\n", stdout);
    }
    /* Output format: Naked JSON */
    else {
        json_printf_dict(stdout, h, 0);
        fputc('\n', stdout);
    }
    g_hash_table_destroy(h);
}
